/*
  Copyright (C) 2010 Julien Jorge

  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the
  Free Software Foundation; either version 2 of the License, or (at your
  option) any later version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
  more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

  contact: julien.jorge@gamned.org
*/
#include <iostream>
#include <fstream>
#include <list>
#include <vector>

#include "kp.hpp"

/**
 * \brief Un tat de la programmation dynamique.
 */
struct state
{
  /** Le meilleur profit pouvant tre obtenu  cet tat. */
  unsigned int profit;

  /** La capacit restante dans le sac. */
  unsigned int remaining_capacity;

  state( unsigned int p, unsigned int r )
    : profit(p), remaining_capacity(r)
  { }
};

/**
 * \brief Cette fonction construit la liste des tats obtenus en considrant la
 *        slection ou la nonslection d'un objet  partir d'une liste d'tats
 *        donne. Le rsultat est une liste d'tats dont le profit est le
 *        maximum obtenu selon la slection.
 * \param data La donne de l'objet considr.
 * \param state_list (entre/sortie) La liste des tats  tendre.
 */
void add_item( const kp_item& data, std::list<state>& state_list )
{
  std::list<state> new_list;
  std::list<state>::iterator it;
  
  // on cre une nouvelle liste o l'objet est ajout
  for (it=state_list.begin(); it!=state_list.end(); ++it)
    if (it->remaining_capacity >= data.weight )
      new_list.push_back
        ( state
          ( it->profit + data.profit, it->remaining_capacity - data.weight ) );
    else
      break;

  // puis on fusionne la nouvelle liste et l'ancienne en gardant l'tat avec le
  // plus grand profit quand deux tats ont la mme capacit restante.
  // Nous commenons par stocker la liste initiale dans old_list et par vider
  // state_list pour la prparer  recevoir les tats rsultant de la fusion.
  std::list<state> old_list;
  std::swap(old_list, state_list);

  /**
   * Nous allons parcourir en parallle la nouvelle liste et l'ancienne. Elles
   * sont ordonnes selon la capacit restante dcroissante et il en sera de
   * mme pour la liste retourne. Par consquent, trois cas se prsentent
   * durant le parcours :
   *
   * # Les tats en tte des deux listes ont la mme capacit restante. Nous
   *   construisons donc un nouvel tat avec la mme capacit et un profit gal
   *   au max des profits des deux tats. Nous continuons ensuite le traitement
   *   sur les restes des deux listes.
   * # L'tat en tte de la premire liste a une capacit restante suprieure 
   *   celle de l'tat en tte de la seconde liste. Nous ajoutons donc ce
   *   premier tat  la liste rsultat puis nous fusionnons le reste de la
   *   premire liste avec toute la seconde liste.
   * # L'tat en tte de la seconde liste a une capacit restante suprieure 
   *   celle de l'tat en tte de la premire liste. Nous ajoutons donc ce
   *   premier tat  la liste rsultat puis nous fusionnons toute la premire
   *   liste avec le reste de la seconde liste.
   *
   * La variable best est mise  jour durant le parcours avec le plus grand
   * profit obtenu depuis le dbut du traitement. Si un nouvel tat a un profit
   * infrieur  cette valeur, il est ignor. En effet, comme les tats sont
   * ordonns par capacit restante dcroissante, il existe ncessairement un
   * autre tat avec un profit meilleur (gal  best) et un plus grand espace
   * disponible pour les ojets restants.
   */
  unsigned int best(0);

  while ( !old_list.empty() && !new_list.empty() )
    {
      state new_state(0, 0);

      if ( old_list.front().remaining_capacity
           == new_list.front().remaining_capacity )
        {
          new_state =
            state
            ( std::max(old_list.front().profit, new_list.front().profit),
              old_list.front().remaining_capacity);
          old_list.pop_front();
          new_list.pop_front();
        }
      else if ( old_list.front().remaining_capacity
                > new_list.front().remaining_capacity )
        {
          new_state = old_list.front();
          old_list.pop_front();
        }
      else
        {
          new_state = new_list.front();
          new_list.pop_front();
        }

      if ( new_state.profit >= best )
        {
          best = new_state.profit;
          state_list.push_back(new_state);
        }
    }

  // une seule des deux boucles ci-dessous sera effectue, au plus
  for ( ; !old_list.empty(); old_list.pop_front() )
    if ( old_list.front().profit >= best )
      {
        best = old_list.front().profit;
        state_list.push_back(old_list.front());
      }

  for ( ; !new_list.empty(); new_list.pop_front() )
    if ( new_list.front().profit >= best )
      {
        best = new_list.front().profit;
        state_list.push_back(new_list.front());
      }
}

/**
  * \brief Rsolution d'un problme de sac  dos.
  * \param data Les donnes des objets.
  * \param W La capacit du sac.
  * \return La valeur des solutions optimales.
  */
unsigned int dp_kp
( const std::vector<kp_item>& data, unsigned int W )
{
  // La liste des tats.
  std::list<state> L;
  L.push_back( state(0, W) );

  // Nous ajoutons les objets un  un.
  for (std::size_t i=0; i!=data.size(); ++i)
    add_item( data[i], L );

  // Avec le filtrage des tats effectu dans add_item(), le dernier tat a
  // ncessairement le plus grand profit.
  return L.back().profit;
}

int main( int argc, char* argv[] )
{
  if ( argc != 2 )
    {
      std::cout << argv[0] << " instance" << std::endl;
      return 1;
    }

  std::ifstream f(argv[1]);
  unsigned int W; // capacit du sac
  std::vector<kp_item> data;
  
  f >> W;

  unsigned int val_p, val_w;

  while (f >> val_p >> val_w)
    data.push_back( kp_item(val_p, val_w) );

  std::cout << "result=" << dp_kp(data, W) << std::endl;

  return 0;
}

