#ifndef __KP_DP_HPP__
#define __KP_DP_HPP__

#include "kp.hpp"

/**
 * \brief Un tat de la programmation dynamique.
 * \param Profit Le meilleur profit pouvant tre obtenu  cet tat.
 * \param RemainingCapacity La capacit restante dans le sac.
 */
template<unsigned int Profit, unsigned int RemainingCapacity>
struct state
{
  static const unsigned int profit = Profit;
  static const unsigned int remaining_capacity = RemainingCapacity;
};

/**
 * \brief Cette structure ajoute un objet dans tous les tats d'une liste
 *        d'tats. Le rsultat est une liste d'tats reprsente par une
 *        dfinition de type nomme result.
 * \param Item L'objet ajout.
 * \param StateList La liste des tats.
 *
 * En notant StateList = ( (p1, w1), (p2, w2), ..., (pk, wk) ), cette structure
 * construit la liste des tats (pi+Item::profit, wi-Item::weight) pour tout i
 * tel que wi soit suprieur  Item::weight.
 */
template<typename Item, typename StateList>
struct add_item_to_all;

/**
 * \brief Spcialisation pour une liste vide, le rsultat est une liste vide.
 */
template<typename Item>
struct add_item_to_all<Item, empty_list>
{
  typedef empty_list result;
};

/* Dclaration complte. */
template<typename Item, typename Head, typename Tail>
struct add_item_to_all< Item, type_list<Head, Tail> >
{
  typedef typename
  if_then_else
  <
    is_less_equal<Item::weight, Head::remaining_capacity>::result,
    type_list
    <
      // On cre l'tat obtenu  partir du premier de StateList.
      state< Head::profit + Item::profit,
             Head::remaining_capacity - Item::weight>,

      // puis on recommence le traitement sur le reste de la liste.
      typename add_item_to_all
      <
        Item,
        Tail
      >::result
    >,
    empty_list // la liste des tats est ordonne selon la capacit restante
            // dcroissante. Par consquent, si cet objet ne peut tre ajout 
            // un tat, il est inutile d'essayer de l'ajouter aux suivants.
  >::result result;
};

/**
 * \brief Cette structure fusionne des listes d'tats en gardant le max des
 *        profits pour une capacit donne. Le rsultat est une liste d'tats
 *        reprsente par une dfinition de type nomme result.
 * \param StateList1 La premire liste,  fusionner avec la seconde.
 * \param StateList2 La seconde liste,  fusionner avec la premire.
 */
template<typename StateList1, typename StateList2>
struct merge_state_lists;

/**
 * \brief Spcialisation pour le cas o la premire liste est vide. Le rsultat
 *        est la seconde liste.
 */
template<typename StateList>
struct merge_state_lists<empty_list, StateList>
{
  typedef StateList result;
};

/**
 * \brief Spcialisation pour le cas o la seconde liste est vide. Le rsultat
 *        est la premire liste.
 */
template<typename StateList>
struct merge_state_lists<StateList, empty_list>
{
  typedef StateList result;
};

/**
 * \brief Spcialisation pour le cas o les deux listes sont vides, pour lever
 *        l'ambiguit des spcialisation prcdentes. Le rsultat est une liste
 *        vide.
 */
template<>
struct merge_state_lists<empty_list, empty_list>
{
  typedef empty_list result;
};

/* Dclaration complte */
/**
 * Nous allons parcourir en parallle les listes passes en paramtres. Elles
 * sont ordonnes selon la capacit restante dcroissante et il en sera de mme
 * pour la liste retourne. Par consquent, trois cas se prsentent durant le
 * parcours :
 *
 * # Les tats en tte des deux listes ont la mme capacit restante. Nous
 *   construisons donc un nouvel tat avec la mme capacit et un profit gal au
 *   max des profits des deux tats. Nous continuons ensuite le traitement sur
 *   les restes des deux listes.
 * # L'tat en tte de la premire liste a une capacit restante suprieure 
 *   celle de l'tat en tte de la seconde liste. Nous ajoutons donc ce premier
 *   tat  la liste rsultat puis nous fusionnons le reste de la premire liste
 *   avec toute la seconde liste.
 * # L'tat en tte de la seconde liste a une capacit restante suprieure 
 *   celle de l'tat en tte de la premire liste. Nous ajoutons donc ce premier
 *   tat  la liste rsultat puis nous fusionnons toute la premire liste avec
 *   le reste de la seconde liste.
 */
template<typename Head1, typename Tail1, typename Head2, typename Tail2>
struct merge_state_lists< type_list<Head1, Tail1>, type_list<Head2, Tail2> >
{
  typedef typename
  if_then_else
  <
    is_equal<Head1::remaining_capacity, Head2::remaining_capacity>::result,
    type_list
    <
      state
      <
        max_value<Head1::profit, Head2::profit>::result,
        Head1::remaining_capacity
      >,
      typename merge_state_lists<Tail1, Tail2>::result
    >,
    typename if_then_else
    <
      is_less<Head2::remaining_capacity, Head1::remaining_capacity>::result,
      type_list
      <
        Head1,
        typename merge_state_lists< Tail1, type_list<Head2, Tail2> >::result
      >,
      type_list
      <
        Head2,
        typename merge_state_lists< type_list<Head1, Tail1>, Tail2 >::result
      >
    >::result
  >::result result;
};

/**
 * \brief Cre une liste ou les tats les plus mauvais sont supprims. Le
 *        rsultat est une liste d'tats reprsente par une dfinition de
 *        type nomme result.
 * \param Best La meilleure valeur connue.
 * \param StateList La liste  traiter.
 */
template<unsigned int Best, typename StateList>
struct clear_list;

/**
 * \brief Spcialisation pour le cas o la liste est vide.
 */
template<unsigned int Best>
struct clear_list<Best, empty_list>
{
  typedef empty_list result;
};

/* Dclaration complte */
/**
 * L'ide est la suivante : si la liste contient deux tats s(p1, w1) et
 * s(p2, w2) tels que w1 > w2 et p1 > p2, alors le premier est clairement
 * meilleur que le second. En effet, il laisse plus d'espace disponible et a un
 * profit plus grand.
 *
 * La liste est initialement ordonne par ordre dcroissant de capacit
 * restante, la slection est alors aise.
 */
template<unsigned int Best, typename Head, typename Tail>
struct clear_list< Best, type_list<Head, Tail> >
{
  typedef typename
  if_then_else
  <
    is_less<Head::profit, Best>::result,
    typename clear_list<Best, Tail>::result,
    type_list<Head, typename clear_list<Head::profit, Tail>::result>
  >::result result;
};

/**
 * \brief Cette structure construit la liste des tats obtenus en considrant la
 *        slection ou la non slection d'un objet  partir d'une liste d'tats
 *        donne. Le rsultat est une liste d'tats dont le profit est le
 *        maximum obtenu selon la slection, reprsente par une dfinition de
 *        type nomme result.
 * \param Item L'objet considr.
 * \param StateList La liste des tats  tendre.
 *
 * La liste donne en paramtre n'est pas modifie lorsque l'objet n'est pas
 * slectionn. Nous construisons la liste obtenue par l'ajout de l'objet en
 * utilisant add_item_to_all, puis nous fusionnons le rsultat avec la liste
 * initiale en utilisant merge_state_lists.
 */
template<typename Item, typename StateList>
struct add_item
{
  typedef typename
  clear_list
  <
    0,
    typename merge_state_lists
    <
      StateList,
      typename add_item_to_all<Item, StateList>::result
    >::result
  >::result result;
};

/**
 * \brief tend une liste d'tats en ajoutant itrativement les objets d'une
 *        liste donne. Le rsultat est une dfinition de type nomme result.
 * \param StateList Les tats  tendre.
 * \param ItemList Les objets  traiter.
 */
template<typename StateList, typename ItemList>
struct dp_kp_loop;

/**
 * \brief Spcialisation pour le cas o il n'y a aucun objet  traiter. Le
 *        rsultat est la liste d'tats donne en paramtre.
 * \param StateList Les tats  tendre.
 */
template<typename StateList>
struct dp_kp_loop<StateList, empty_list>
{
  typedef StateList result;
};

/* Dclaration complte */
/**
 * Le rsultat est obtenu en ajoutant les objets de ItemList::tail dans la liste
 * des tats obtenus par l'ajout de ItemList::head  partir de
 * StateList. L'ajout des objets de ItemList::tail se fait par une utilisation
 * rpte de ce patron de classe.
 */
template<typename StateList, typename ItemHead, typename ItemTail>
struct dp_kp_loop< StateList, type_list<ItemHead, ItemTail> >
{
  typedef typename
  dp_kp_loop
  <
    typename add_item
    <
      ItemHead,
      StateList
    >::result,
    ItemTail
  >::result result;
};

/**
 * \brief Une classe qui rsout un problme de sac  dos en utilisant
 *        l'algorithme de programmation dynamique par listes. Le rsultat est
 *        une dclaration de type nomme result, renommant une instanciation du
 *        patron de classe state<>.
 * \param Capacity La capacit du le sac.
 * \param ItemList Les objets sur lesquels nous faisons une slection.
 *
 * Le rsultat est obtenu en ajoutant les objets de ItemList dans une liste
 * initiale d'tats. Cette liste ne contient qu'un tat o le profit est nul et
 * la capacit disponible est Capacity. L'ajout des objets se fait via le patron
 * de classe add_items<>, dfinissant une liste d'tats atteignables avec tous
 * les objets de ItemList. Par construction, le dernier tat de cette liste a le
 * plus grand profit. Il est slectionn pour dfinir le rsultat.
 */
template<unsigned int Capacity, typename ItemList>
struct dp_kp
{
  typedef typename
  last_type
  <
    typename dp_kp_loop
    <
      // la liste initiale d'tats contient un unique tat ayant un profit nul
      // et dans lequel toute la capacit est disponible. Nous partons de cet
      // tat pour ajouter...
      type_list<state<0, Capacity>, empty_list>,
      // ...tous les objets.
      ItemList
    >::result
  >::result result;
};

#endif // __KP_DP_HPP__
