/*
  Copyright (C) 2010 Julien Jorge

  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the
  Free Software Foundation; either version 2 of the License, or (at your
  option) any later version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
  more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

  contact: julien.jorge@gamned.org
*/
#include <iostream>
#include <fstream>
#include <algorithm>
#include <vector>

#include "kp.hpp"

/**
 * \brief Calcul de la valeur optimale d'un problme de sac  dos, par
 *        application directe des quation de Bellman.
 * \param data Les donnes des objets.
 * \param d La capacit du sac.
 * \param n Le nombre d'objets considrs.
 */
unsigned int solve_kp
( const std::vector<kp_item>& data, unsigned int d, unsigned int n )
{
  if (n==0)    // pas d'objet ?
    return 0;  // la solution est vide
  else
    {
      // Nous allons considrer le dernier objet, dont les donnes sont dans la
      // case n-1 des vecteurs
      const unsigned int i(n-1);

      // Si l'objet i ne rentre pas dans le sac, il ne peut pas tre choisi. La
      // valeur de la solution est donc obtenue en considrant les autres
      // objets, en gardant la mme capacit.
      if (data[i].weight > d)
        return solve_kp(data, d, n-1);
      else
        // Si l'objet i tient dans le sac, il y a deux possibilits. Soit il
        // participe  la solution d'une valeur p[i], auquel cas il ne
        // reste qu'une capacit de d - w[i] pour les autres objets. Soit il ne
        // participe pas  la solution, auquel cas la capacit est
        // inchange. La valeur de la solution optimale est obtenue en gardant
        // la meilleure de ces deux solutions.
        return std::max
          ( solve_kp(data, d-data[i].weight, n-1) + data[i].profit,
            solve_kp(data, d, n-1) );
    }
}

int main( int argc, char* argv[] )
{
  if ( argc != 2 )
    {
      std::cout << argv[0] << " instance" << std::endl;
      return 1;
    }

  std::ifstream f(argv[1]);
  unsigned int W;              // capacit du sac
  std::vector<kp_item> data;   // les donnes des objets

  f >> W;

  // le profit et le poids d'un objet, lu dans le fichier
  unsigned int temp_p, temp_w;

  while (f >> temp_p >> temp_w)
    data.push_back( kp_item(temp_p, temp_w) );

  std::cout << "result=" << solve_kp(data, W, data.size())
            << std::endl;

  return 0;
}
